# paraphase
# Author: Xiao Chen <xchen@pacificbiosciences.com>


import networkx as nx
import matplotlib.pyplot as plt
import copy
from pprint import pprint
import math
import numpy as np
from collections import namedtuple


class VariantGraph:
    HaplotypesReads = namedtuple("HaplotypesReads", "unique nonunique by_read")

    def __init__(self, reads, pivot_pos, figure_id="fig"):
        self.nvar = len(list(reads.values())[0])
        if self.nvar <= 1:
            raise Exception("Too few variants")
        self.reads = reads
        self.reads_original = copy.deepcopy(reads)
        self.pivot_pos = pivot_pos
        self.figure_id = figure_id
        # these three get updated throughout
        self.edge_per_position = {}
        self.nodes = []
        self.edges = []
        # only used when first constructing edges
        self.next_per_node = {}
        self.previous_per_node = {}
        # positions to plot nodes
        self.node_positions = {}
        self.edge_info = {}
        self.dnhap = {}

    def add_edge(self, a, b):
        """Add edges between two nodes"""
        self.edges.append([a, b])

    def rm_edge(self, a, b):
        """Remove edges between two nodes"""
        if [a, b] in self.edges:
            self.edges.remove([a, b])

    def get_positions(self):
        """Get a dictionary of the positions of each node in the graph plot"""
        for i in [1, 2]:
            for j in range(self.nvar):
                self.node_positions.setdefault(f"{i}-{j}", (j * 1000, i * 10))
        for n in self.nodes:
            i, j = n.split("-")
            if len(i) == 1 and f"{i}-{j}" not in self.node_positions:
                i = int(i)
                j = int(j)
                self.node_positions.setdefault(f"{i}-{j}", (j * 1000, i * 10))

    def label_pos_offset(self):
        """Add offset to labels in the graph plot"""
        pos_with_offset = {}
        y_off = 0.5  # offset on the y axis
        for k in self.node_positions.keys():
            v = self.node_positions[k]
            if int(k.split("-")[0]) == 1:
                pos_with_offset[k] = (v[0], v[1] - y_off)
            else:
                pos_with_offset[k] = (v[0], v[1] + y_off)
        return pos_with_offset

    def visualize(self, ax):
        """Make a graph plot"""
        G = nx.DiGraph()
        for n in self.nodes:
            G.add_node(n)
        G.add_edges_from(self.edges)
        nx.draw_networkx(
            G,
            pos=self.node_positions,
            with_labels=False,
            node_size=40,
            node_color="red",
            ax=ax,
            arrows=False,
        )
        label_pos = self.label_pos_offset()
        text = nx.draw_networkx_labels(
            G, pos=label_pos, font_size=25, ax=ax, clip_on=False
        )
        for _, t in text.items():
            t.set_rotation(25)
        plt.plot()

    def path(self, pos1, pos2):
        """Get simple path between two nodes"""
        G = nx.DiGraph()
        G.add_edges_from(self.edges)
        sources = set([a[0] for a in self.edge_per_position[pos1]])
        targets = set([a[1] for a in self.edge_per_position[pos2 - 1]])
        paths = []
        for source in sources:
            for target in targets:
                kk = list(
                    nx.all_simple_edge_paths(G, source, target, cutoff=pos2 - pos1)
                )
                for path in kk:
                    hap = path[0][0].split("-")[0]
                    for site in path:
                        hap += site[1].split("-")[0]
                    paths.append("".join(hap))
        return list(set(paths))

    def plotfig(self):
        """Create figure"""
        figure = plt.gcf()
        figure.set_size_inches(60, 30)
        plt.savefig(f"{self.figure_id}.png")
        plt.close(figure)

    def get_hap_by_pos(self, pos):
        """Given a position, return all the haplotypes starting there"""
        haps = []
        for node in self.nodes:
            if int(node.split("-")[1]) == pos:
                haps.append(node.split("-")[0])
        return haps

    def get_nstart_nend(self):
        """Get the smallest and biggest position numbers in the graph"""
        pos_left = sorted(list(set([int(a.split("-")[1]) for a in self.nodes])))
        if pos_left == []:
            return 0, 0
        nstart = min(pos_left)
        nend = max(pos_left)
        return nstart, nend

    def get_next_pos(self, pos):
        """Get next position and its haplotype blocks in the graph"""
        positions = sorted(list(set([int(a.split("-")[1]) for a in self.nodes])))
        if pos >= positions[-1]:
            return None, None
        for position in positions:
            if position > pos:
                next_pos = position
                break
        next_haps = self.get_hap_by_pos(next_pos)
        return next_pos, next_haps

    def get_previous_pos(self, pos):
        """Get previous position and its haplotype blocks in the graph"""
        positions = sorted(list(set([int(a.split("-")[1]) for a in self.nodes])))
        if pos <= 0:
            return None, None
        for position in positions[::-1]:
            if position < pos:
                prev_pos = position
                break
        prev_haps = self.get_hap_by_pos(prev_pos)
        return prev_pos, prev_haps

    def get_highest_cn(self, final_haps):
        """Return the highest possible copy number given the assembled haplotypes"""
        main_haps_candidates = []
        for i in reversed(range(self.nvar)):
            lhaps = [a for a in final_haps if a[i] != "x"]
            if lhaps not in main_haps_candidates:
                main_haps_candidates.append(lhaps)
        main_haps_candidates = sorted(
            main_haps_candidates, key=lambda x: len(x), reverse=True
        )
        main_haps = main_haps_candidates[0]
        highest_cn = len(main_haps)
        return (
            highest_cn,
            main_haps_candidates,
        )

    def filter_low_support_haps(self, haps_to_assess, min_count=3, debug=False):
        """Remove haplotypes that have a low number of supporting reads"""
        reads = self.reads_original
        read_support = self.match_reads_and_haplotypes(reads, haps_to_assess)
        good_reads = read_support.unique
        filtered_ass_haps = []
        for test_hap in good_reads:
            nread = len(good_reads[test_hap])
            if nread > 0:
                filtered_ass_haps.append(test_hap)
            elif debug:
                print(f"removing low support hap {test_hap} {nread}")
        haps_to_assess = filtered_ass_haps
        while True:
            read_support = self.match_reads_and_haplotypes(reads, haps_to_assess)
            good_reads = read_support.unique
            filtered_ass_haps = []
            for test_hap in good_reads:
                nread = len(good_reads[test_hap])
                if nread >= min_count:
                    filtered_ass_haps.append(test_hap)
                elif debug:
                    print(f"removing low support hap {test_hap} {nread}")
            if len(filtered_ass_haps) == len(haps_to_assess):
                break
            else:
                haps_to_assess = filtered_ass_haps
        if debug:
            print(
                "All removed haplotypes are ",
                set(haps_to_assess) - set(filtered_ass_haps),
            )
        return filtered_ass_haps

    @staticmethod
    def get_thres(num_reads):
        """Get thresholds for pruning some noisy links"""
        min_num = min(num_reads) if num_reads != [] else 0
        remain = [a for a in num_reads if a != min_num]
        if remain != []:
            if min(remain) > 10:
                return min(5, max(2, math.floor(min(remain) / 25)))
            if min(remain) > 5:
                return 1
        return 0

    # update the edges
    def get_edges(self, debug=False):
        """Create and filter edges"""
        dedge = {}
        reads = self.reads
        for read in reads:
            hap = reads[read]
            nvar = self.nvar
            if hap.count("x") < nvar - 1:
                for i in range(nvar - 1):
                    base1 = hap[i]
                    base2 = hap[i + 1]
                    if "x" not in [base1, base2]:
                        node1 = base1 + "-" + str(i)
                        node2 = base2 + "-" + str(i + 1)
                        dedge.setdefault((node1, node2), []).append(read)
        for node1, node2 in dedge:
            pos = int(node1[2:])
            self.edge_per_position.setdefault(pos, []).append(
                (node1, node2, len(dedge[(node1, node2)]))
            )
            self.next_per_node.setdefault(node1, []).append(node2)
            self.previous_per_node.setdefault(node2, []).append(node1)
        new_edge_per_position = {}
        for i in range(self.nvar - 1):
            new_edge_per_position.setdefault(i, [])

        # only remove those edges that have low support and won't cause an
        # orphan node if removed
        edges_removed = []
        for pos in self.edge_per_position:
            num_reads = [at[2] for at in self.edge_per_position[pos]]
            # when variants are inside a deletion region
            has_sv = True in [
                at[0].split("-")[0] not in ["1", "2"]
                or at[1].split("-")[0] not in ["1", "2"]
                for at in self.edge_per_position[pos]
            ]
            thres = self.get_thres(num_reads)
            for (node1, node2, n) in self.edge_per_position[pos]:
                if (
                    len(self.next_per_node[node1]) > 1
                    and len(self.previous_per_node[node2]) > 1
                    and n <= thres
                ) or (has_sv and n == 1):
                    edges_removed.append((node1, node2, n))
                else:
                    new_edge_per_position.setdefault(pos, []).append((node1, node2, n))
        self.edge_per_position = new_edge_per_position
        for (node1, node2, n) in edges_removed:
            if debug:
                print("removing", node1, node2, n)
            self.next_per_node[node1].remove(node2)
            self.previous_per_node[node2].remove(node1)
        for pos in self.edge_per_position:
            num_reads = [at[2] for at in self.edge_per_position[pos]]
            self.edge_info.setdefault(pos, num_reads)

    def summerize_edges(self):
        """Check the number of different edges starting at each position"""
        edge_per_position = self.edge_per_position
        for pos in edge_per_position:
            nhap = -1
            if (
                pos in self.edge_info
                and self.edge_info[pos] != []
                and min(self.edge_info[pos]) >= 2
                and max(self.edge_info[pos]) > 2
            ):
                dlinks = {}
                site1 = set()
                site2 = set()
                expected_site1 = set()
                expected_site2 = set()
                for node in self.nodes:
                    base, base_pos = node.split("-")
                    base_pos = int(base_pos)
                    if base_pos == pos:
                        expected_site1.add(base)
                    elif base_pos == pos + 1:
                        expected_site2.add(base)
                for node1, node2, _ in edge_per_position[pos]:
                    site1.add(node1.split("-")[0])
                    site2.add(node2.split("-")[0])
                    dlinks.setdefault(node1, set()).add(node2)
                # add some test here
                if sorted(site1) == sorted(expected_site1) and sorted(site2) == sorted(
                    expected_site2
                ):
                    if [a for a in dlinks if len(dlinks[a]) > 1] == []:
                        nhap = 2
                    else:
                        nhap = 10
            self.dnhap.setdefault(pos, nhap)
        # break the graph at the pivot pos, can make it optional
        if self.pivot_pos - 1 in self.dnhap:
            if self.dnhap[self.pivot_pos - 1] == 2:
                self.dnhap[self.pivot_pos - 1] = 10

    def get_segments(self):
        """Break graph into segments"""
        segments = {}
        dnhap = dict(sorted(self.dnhap.items()))
        dnhap_bk = copy.deepcopy(self.dnhap)
        list_keys = list(dnhap_bk.keys())
        list_values = list(dnhap_bk.values())
        if list_values[:3] == [10, 2, 2] and list_values != [10, 2, 2, 10]:
            dnhap[list_keys[0]] = 2
        if list_values[-3:] == [2, 2, 10] and list_values != [10, 2, 2, 10]:
            dnhap[list_keys[-1]] = 2
        for i, val in enumerate(list_values):
            if i > 0 and i < len(list_values) - 1:
                if list_values[i - 1] == 10 and val == 2 and list_values[i + 1] == 10:
                    dnhap[list_keys[i]] = 10

        lpos = [0]
        last_nhap = dnhap[0]
        for pos in dnhap:
            if pos > 0:
                nhap = dnhap[pos]
                if nhap != last_nhap:
                    segments.setdefault((min(lpos), max(lpos)), last_nhap)
                    lpos = [pos]
                    last_nhap = nhap
                else:
                    lpos.append(pos)
        segments.setdefault((min(lpos), max(lpos)), last_nhap)
        return segments

    def update_reads(self):
        """
        After pruning edges, update the corresponding sites in reads to "x"
        """
        reads = self.reads
        edge_per_position = self.edge_per_position
        new_reads = {}
        for read in reads:
            hap = reads[read]
            new_hap = list(hap)
            nvar = len(hap)
            for i in range(nvar - 1):
                base1 = hap[i]
                base2 = hap[i + 1]
                if "x" not in [base1, base2]:
                    found = False
                    for node1, node2, n in edge_per_position[i]:
                        if node1 == base1 + "-" + str(i) and node2 == base2 + "-" + str(
                            i + 1
                        ):
                            found = True
                    if found is False:
                        new_hap[i] = "x"
                        new_hap[i + 1] = "x"
            if new_hap.count("x") <= nvar - 2:
                new_reads.setdefault(read, "".join(new_hap))
        self.reads = new_reads

    def merge_two_pos(self, new_nodes, debug=False):
        """
        Parameters:
            new_nodes (dict): two positions and their associated partial haplotypes
        Returns:
            sub_hap_support_reads (dict): dict of (merged haps):(supporting reads)
            dnext (dict): dict of (hap1):(set of potential linking hap2s downstream)
            dbefore (dict): dict of (hap1):(set of potential linking hap2s upstream)
        """
        reads = self.reads
        sub_hap_support_reads = {}
        dnext = {}
        dbefore = {}
        for read in reads:
            hap = reads[read]
            best_matches = []
            nmatches = []
            for pos in new_nodes:
                _, best_match, max_match = self.get_matching_hap(new_nodes, pos, hap)
                nmatches.append(max_match)
                best_matches.append(best_match)
            if None not in best_matches:
                # if debug:
                #    print(read, hap, best_matches, nmatches)
                num_matches = [len(a) for a in best_matches]
                if num_matches == [1, 1]:
                    joined_haplotype = best_matches[0][0] + "" + best_matches[1][0]
                    sub_hap_support_reads.setdefault(joined_haplotype, []).append(read)
                for a in best_matches[0]:
                    for b in best_matches[1]:
                        dnext.setdefault(a, set()).add(b)
                        dbefore.setdefault(b, set()).add(a)
        return sub_hap_support_reads, dnext, dbefore

    @staticmethod
    def get_missing(subregion_haps_assembled, left_original, right_original):
        """
        find those that cannot span
        Parameters:
            subregion_haps_assembled (lst): assembled haps
            left_original (lst): left candidates
            right_original (lst): right candidates
        Returns:
            left_missing (lst): left candidates missing
            right_missing (lst): right candidates missing
        """
        left_missing = []
        right_missing = []
        for test_hap in left_original:
            found = False
            for ass_hap in subregion_haps_assembled:
                if test_hap == ass_hap[: len(test_hap)]:
                    found = True
            if found is False:
                left_missing.append(test_hap)
        for test_hap in right_original:
            found = False
            for ass_hap in subregion_haps_assembled:
                if test_hap == ass_hap[0 - len(test_hap) :]:
                    found = True
            if found is False:
                right_missing.append(test_hap)
        return left_missing, right_missing

    # get haplotypes between pos1 and pos2
    def subregion_assembly(self, pos1, pos2, allow_x=False, debug=False):
        """
        Assemble the region between pos1 and pos2
        Parameters:
            allow_x (bool): if true, missing candidates will be output with x's appended
        Returns:
            bool: assembly success
            pos1_haps (lst of str): assembled haps
            dnext (dict):
            dbefore (dict):
        """
        pos1_next = pos1
        pos1_haps = self.get_hap_by_pos(pos1)

        # extend positions one by one
        while pos1_next < pos2:
            next_pos, next_haps = self.get_next_pos(pos1_next)
            assert pos1_haps != [] and next_haps != []
            subregion_haps_assembled = set()

            if len(pos1_haps) == 1 or len(next_haps) == 1:
                for a in pos1_haps:
                    for b in next_haps:
                        subregion_haps_assembled.add(a + b)
                        dnext = {}
                        dbefore = {}
            else:
                new_nodes = {pos1: pos1_haps, next_pos: next_haps}
                if debug:
                    print("positions", pos1, next_pos)
                    print(new_nodes)
                sub_hap_support_reads, dnext, dbefore = self.merge_two_pos(
                    new_nodes, debug=debug
                )

                # do some filtering
                thres = 2  # 3
                hap_len1 = len(pos1_haps[0])
                hap_len2 = len(next_haps[0])
                hap_lens = [hap_len1, hap_len2]
                if min(hap_lens) > 2 or (min(hap_lens) == 2 and max(hap_lens) >= 10):
                    thres = 1  # 2
                for sub_hap in sub_hap_support_reads:
                    if debug:
                        print(sub_hap, len(sub_hap_support_reads[sub_hap]))
                    if len(sub_hap_support_reads[sub_hap]) >= thres:
                        subregion_haps_assembled.add(sub_hap)

            left_missing, right_missing = self.get_missing(
                subregion_haps_assembled, pos1_haps, next_haps
            )

            if right_missing != [] or left_missing != []:
                if allow_x is False:
                    return False, pos1_haps, dnext, dbefore

                if debug:
                    pprint(dnext)
                # rescue missing haps
                if (
                    hap_len1 >= 3
                    and hap_len2 >= 3
                    and (next_pos - 1) in self.dnhap
                    and self.dnhap[next_pos - 1] != -1
                    and (next_pos - 1) in self.edge_info
                    and self.edge_info[next_pos - 1] != []
                    and min(self.edge_info[next_pos - 1]) >= 2
                ):
                    if debug:
                        print("missing", left_missing, right_missing)
                    rescued_haps = self.rescue_missing(
                        left_missing, right_missing, dnext, dbefore, debug=debug
                    )
                    subregion_haps_assembled = subregion_haps_assembled.union(
                        rescued_haps
                    )

                    left_missing, right_missing = self.get_missing(
                        subregion_haps_assembled, pos1_haps, next_haps
                    )
                    if debug:
                        print("missing2", left_missing, right_missing)

                # append x's to missing haps
                if right_missing != []:
                    len_x = len(pos1_haps[0])
                    for test_hap in right_missing:
                        subregion_haps_assembled.add("x" * len_x + test_hap)
                if left_missing != []:
                    len_x = len(next_haps[0])
                    for test_hap in left_missing:
                        subregion_haps_assembled.add(test_hap + "x" * len_x)

            if debug:
                print("after merging ", subregion_haps_assembled)
            pos1_haps = list(subregion_haps_assembled)
            pos1_next = next_pos

        return True, pos1_haps, dnext, dbefore

    @staticmethod
    def rescue_missing(left_missing, right_missing, dnext, dbefore, debug=False):
        """
        Two scenarios:
        1. there is only one link and that one link is the only one in the other
        missing set, or the other missing set is empty
        2. there can be multiple links but only one of them is in the other missing
        set, and that one link is the only one in the other missing set
        """
        rescued_haps = set()
        for missing_hap in left_missing:
            if missing_hap in dnext and missing_hap.count("x") <= 1:
                possible_links = list(dnext[missing_hap])
                if len(possible_links) == 1 and right_missing in [
                    [],
                    [possible_links[0]],
                ]:
                    if debug:
                        print("merging", missing_hap, possible_links)
                    rescued_haps.add(missing_hap + possible_links[0])
                else:
                    possible_links_also_missing = [
                        a for a in possible_links if a in right_missing
                    ]
                    if (
                        len(possible_links_also_missing) == 1
                        and right_missing == possible_links_also_missing
                        and len(left_missing) == 1
                    ):
                        if debug:
                            print(
                                "merging",
                                missing_hap,
                                possible_links,
                                possible_links_also_missing,
                            )
                        rescued_haps.add(missing_hap + possible_links_also_missing[0])
        for missing_hap in right_missing:
            if missing_hap in dbefore and missing_hap.count("x") <= 1:
                possible_links = list(dbefore[missing_hap])
                if len(possible_links) == 1 and left_missing in [
                    [],
                    [possible_links[0]],
                ]:
                    if debug:
                        print("merging", possible_links, missing_hap)
                    rescued_haps.add(possible_links[0] + missing_hap)
                else:
                    possible_links_also_missing = [
                        a for a in possible_links if a in left_missing
                    ]
                    if (
                        len(possible_links_also_missing) == 1
                        and left_missing == possible_links_also_missing
                        and len(right_missing) == 1
                    ):
                        if debug:
                            print(
                                "merging",
                                possible_links,
                                possible_links_also_missing,
                                missing_hap,
                            )
                        rescued_haps.add(possible_links_also_missing[0] + missing_hap)
        return rescued_haps

    @staticmethod
    def get_matching_hap(new_nodes, pos1, hap):
        """
        Parameters:
            new_nodes (dict): two positions and their associated partial haplotypes
            pos1 (int): positions of interest
            hap (str): full read bases
        Returns:
            hap1: corresponding segment in the read
            best_match (lst of str): list of best matches
            max_match (int): length of the best matches
        """
        candidates = list(new_nodes[pos1])
        pos2 = len(candidates[0])
        hap1 = hap[pos1 : (pos1 + pos2)]
        matches = {}
        for candidate_hap in candidates:
            match, mismatch, _ = VariantGraph.compare_two_haps(candidate_hap, hap1)
            if mismatch == 0 and match > 0:
                matches.setdefault(match, []).append(candidate_hap)
        best_match = None
        max_match = 0
        if matches != {}:
            all_matches = list(matches.keys())
            max_match = max(all_matches)
            best_match = matches[max_match]
        return hap1, best_match, max_match

    @staticmethod
    def compare_two_haps(hap1, hap2):
        """
        Calculate number of matching/mismatching/additional bases between two haplotypes
        """
        if len(hap1) != len(hap2):
            raise Exception("Two haplotypes are not equal length")
        match = 0
        mismatch = 0
        extra = 0
        for i, base1 in enumerate(hap1):
            base2 = hap2[i]
            if "x" not in [base1, base2]:
                if base1 == base2:
                    match += 1
                else:
                    mismatch += 1
            if base1 == "x" and base2 != "x":
                extra += 1
        return match, mismatch, extra

    @staticmethod
    def match_reads_and_haplotypes(haplotype_per_read, hap_list, min_match=1):
        """
        Get reads that support each haplotype and possible haplotypes of each read
        """
        support_reads = {}
        support_reads_nonunique = {}
        supporting_haps_per_read = {}
        for read_name in haplotype_per_read:
            read_hap = haplotype_per_read[read_name]
            matching_haplotypes = []
            for haplotype_to_extend in hap_list:
                match, mismatch, extend = VariantGraph.compare_two_haps(
                    haplotype_to_extend, read_hap
                )
                if mismatch == 0 and match >= min_match and extend >= 0:
                    matching_haplotypes.append(haplotype_to_extend)
                    support_reads_nonunique.setdefault(haplotype_to_extend, []).append(
                        read_hap
                    )
            supporting_haps_per_read.setdefault(read_name, matching_haplotypes)
            # unique match
            if len(matching_haplotypes) == 1:
                support_reads.setdefault(matching_haplotypes[0], []).append(read_hap)
            else:
                # allow for scenarios like a read "x11x" supporting both "21xx" and "xx12"
                for hap1 in matching_haplotypes:
                    has_overlap = False
                    for hap2 in matching_haplotypes:
                        if hap1 != hap2:
                            match, mismatch, extend = VariantGraph.compare_two_haps(
                                hap1, hap2
                            )
                            if match != 0 or mismatch != 0:
                                has_overlap = True
                                break
                    # has multiple supporting haplotypes but they have no overlap at all
                    # so all are unique
                    if has_overlap is False:
                        support_reads.setdefault(hap1, []).append(read_hap)
        return VariantGraph.HaplotypesReads(
            support_reads,
            support_reads_nonunique,
            supporting_haps_per_read,
        )

    def rm_add_edges(self, pos1, pos2, new_haps):
        """
        Add new nodes/edges between two positions and remove old edges/nodes
        Updates self.nodes, self.edge_per_position, self.edges
        Also updates self.node_positions for plotting
        """
        new_pos = pos1
        old_edge_per_position = copy.deepcopy(self.edge_per_position)

        prev_pos, _ = self.get_previous_pos(pos1)
        # rm edges, from prev_pos to pos2+1
        nstart = 0
        if prev_pos is not None:
            nstart = prev_pos
        for pos in range(nstart, pos2 + 1):
            if pos in self.edge_per_position:
                for node1, node2, _ in self.edge_per_position[pos]:
                    self.rm_edge(node1, node2)

        # pop positions before and next
        if prev_pos is not None:
            self.edge_per_position.pop(prev_pos, None)
        for n in range(pos1, pos2 + 1):
            self.edge_per_position.pop(n, None)
            old_nodes = [a for a in self.nodes if int(a.split("-")[1]) == n]
            for old_node in old_nodes:
                self.nodes.remove(old_node)

        for i, new_hap in enumerate(new_haps):
            new_node = f"{new_hap}-{pos1}"
            self.nodes.append(new_node)
            if len(new_haps) > 2:
                self.node_positions.setdefault(new_node, (new_pos * 1000, i * 5 + 10))
            else:
                self.node_positions.setdefault(new_node, (new_pos * 1000, (i + 1) * 10))
            # add edges after
            if pos2 in old_edge_per_position:
                old_edges = old_edge_per_position[pos2]
                for node1, node2, _ in old_edges:
                    if new_hap[-1] == node1[0]:
                        self.add_edge(new_node, node2)
                        self.edge_per_position.setdefault(pos1, [])
                        if (new_node, node2, None) not in self.edge_per_position[pos1]:
                            self.edge_per_position[pos1].append((new_node, node2, None))
            # add edges before
            if prev_pos is not None and prev_pos in old_edge_per_position:
                old_edges = old_edge_per_position[prev_pos]
                for node1, node2, _ in old_edges:
                    if new_hap[0] == node2[0]:
                        self.add_edge(node1, new_node)
                        self.edge_per_position.setdefault(prev_pos, [])
                        if (node1, new_node, None) not in self.edge_per_position[
                            prev_pos
                        ]:
                            self.edge_per_position[prev_pos].append(
                                (node1, new_node, None)
                            )

    def merge_edges_simple(self, pos1, pos2):
        """Merge edges when there are two haplotypes only"""
        new_haps = self.path(pos1, pos2)
        self.rm_add_edges(pos1, pos2, new_haps)

    def merge_edges(self, pos1, pos2, debug=False):
        """Merge edges in more complex scenario - more than two haplotypes"""
        success, new_haps, _, _ = self.subregion_assembly(pos1, pos2, debug=debug)
        if success is True:
            self.rm_add_edges(pos1, pos2, new_haps)
        return new_haps

    def initialize_graph(self, debug=False):
        """Initialize graph, add and prune edges"""
        self.get_edges(debug=debug)
        self.update_reads()

        for pos in self.edge_per_position:
            for node1, node2, _ in self.edge_per_position[pos]:
                self.add_edge(node1, node2)
                if node1 not in self.nodes:
                    self.nodes.append(node1)
                if node2 not in self.nodes:
                    self.nodes.append(node2)

        for i in [1, 2]:
            for j in range(self.nvar):
                node = f"{i}-{j}"
                if node not in self.nodes:
                    hap = "x" * j + str(i) + "x" * (self.nvar - j - 1)
                    read_support = self.match_reads_and_haplotypes(
                        self.reads_original, [hap]
                    )
                    if hap in read_support.unique and len(read_support.unique[hap]) > 1:
                        self.nodes.append(node)

        for (node1, node2) in self.edges:
            for n in (node1, node2):
                if n not in self.nodes:
                    self.nodes.append(n)
        self.get_positions()
        if debug:
            pprint(self.edge_per_position)

    def filter_candidates(self, hap_block, candidates, lenx, candidate_first=False):
        """
        Filter candidates against reads
        """
        read_support = {}
        reads = self.reads
        hap1 = hap_block
        for candidate in candidates:
            candidate_len = len(candidate)
            if candidate_first:
                hap2 = (
                    "x" * (lenx - candidate_len) + candidate + "x" * (self.nvar - lenx)
                )
            else:
                hap2 = (
                    "x" * (self.nvar - lenx) + candidate + "x" * (lenx - candidate_len)
                )
            for read in reads:
                read_seq = reads[read]
                match1, mismatch1, _ = VariantGraph.compare_two_haps(hap1, read_seq)
                match2, mismatch2, _ = VariantGraph.compare_two_haps(hap2, read_seq)
                if mismatch1 == 0 and mismatch2 == 0:
                    if match1 > 0 and match2 > 0:
                        read_support.setdefault(candidate, []).append(
                            min(match1, match2)
                        )
        if read_support == {}:
            return set()
        if len(read_support) == 1:
            return [list(read_support.keys())[0]]
        medians = []
        maxs = []
        for candidate in read_support:
            medians.append(np.median(read_support[candidate]))
            maxs.append(max(read_support[candidate]))
        for i in range(len(medians)):
            others = []
            for j in range(len(medians)):
                if j != i:
                    others.append(maxs[j])
            if medians[i] > max(others) and medians[i] >= 4:
                new_candidate = list(read_support.keys())[i]
                if len(read_support[new_candidate]) >= 4:
                    return [new_candidate]
        return candidates

    def extend_pivot_blocks(self, main_haps, debug=False):
        """
        Take candidate matching haplotypes and extend with shared bases
        """
        new_main_haps = []
        for hap in main_haps:
            identical_bases_right = ""
            identical_bases_left = ""
            prefix_x = 0
            suffix_x = self.nvar - 1
            if "x" in hap:
                for prefix_x in range(self.nvar):
                    if hap[prefix_x] != "x":
                        break
                for suffix_x in reversed(range(self.nvar)):
                    if hap[suffix_x] != "x":
                        break
                if debug:
                    print(f"extending {hap} {prefix_x} {suffix_x}")
                # extend right
                if suffix_x < self.nvar - 2:
                    next_pos = suffix_x + 1
                    # check that this is not a site very distant from next site
                    if (
                        suffix_x in self.dnhap
                        and self.dnhap[suffix_x] != -1
                        and suffix_x in self.edge_info
                        and self.edge_info[suffix_x] != []
                        and min(self.edge_info[suffix_x]) >= 2
                    ):
                        previous_pos, _ = self.get_previous_pos(next_pos)
                        _, _, dnext, dbefore = self.subregion_assembly(
                            previous_pos, next_pos, allow_x=False, debug=debug
                        )
                        if debug:
                            pprint(dnext)
                        hap_block = hap[previous_pos : suffix_x + 1]
                        if hap_block in dnext:  # and len(hap_block) >= 5:
                            candidates = dnext[hap_block]
                            # check candidates against reads
                            if len(candidates) > 1:
                                candidates = self.filter_candidates(
                                    hap, candidates, self.nvar - 1 - suffix_x
                                )
                            if candidates != set():
                                next_hap_len = len(list(candidates)[0])
                                i = 0
                                while True:
                                    if i >= next_hap_len:
                                        break
                                    bases_per_candidate = list(
                                        set([test_hap[i] for test_hap in candidates])
                                    )
                                    if len(bases_per_candidate) != 1:
                                        break
                                    identical_bases_right += bases_per_candidate[0]
                                    i += 1
                # extend left
                if prefix_x >= 2:
                    next_pos = prefix_x
                    _, nend = self.get_nstart_nend()
                    if next_pos < nend:
                        next_next_pos, _ = self.get_next_pos(next_pos)
                        hap_block = hap[prefix_x:next_next_pos]
                    else:
                        hap_block = hap[prefix_x:]
                    # check that this is not a site very distant from previous site
                    if (
                        (prefix_x - 1) in self.dnhap
                        and self.dnhap[prefix_x - 1] != -1
                        and (prefix_x - 1) in self.edge_info
                        and self.edge_info[prefix_x - 1] != []
                        and min(self.edge_info[prefix_x - 1]) >= 2
                    ):
                        previous_pos, _ = self.get_previous_pos(next_pos)
                        _, _, dnext, dbefore = self.subregion_assembly(
                            previous_pos, next_pos, allow_x=False, debug=debug
                        )
                        if debug:
                            pprint(dbefore)
                        if hap_block in dbefore:  # and len(hap_block) >= 5:
                            candidates = dbefore[hap_block]
                            # check candidates against reads
                            if len(candidates) > 1:
                                candidates = self.filter_candidates(
                                    hap,
                                    candidates,
                                    prefix_x,
                                    candidate_first=True,
                                )
                            if candidates != set():
                                previous_hap_len = len(list(candidates)[0])
                                i = previous_hap_len - 1
                                while True:
                                    if i < 0:
                                        break
                                    bases_per_candidate = list(
                                        set([test_hap[i] for test_hap in candidates])
                                    )
                                    if len(bases_per_candidate) != 1:
                                        break
                                    identical_bases_left += bases_per_candidate[0]
                                    i -= 1
                                identical_bases_left = identical_bases_left[::-1]
            extended_hap = (
                "x" * (prefix_x - len(identical_bases_left))
                + identical_bases_left
                + hap[prefix_x : suffix_x + 1]
                + identical_bases_right
                + "x" * (self.nvar - 1 - suffix_x - len(identical_bases_right))
            )
            if debug and extended_hap != hap:
                print(f"after extension is {extended_hap}")
            new_main_haps.append(extended_hap)
        return new_main_haps

    def assemble_haplotypes(self, debug=False, make_plot=True):
        """Stepwise assembly of haplotype blocks based on graph"""
        if make_plot:
            _, axs = plt.subplots(5)
            self.visualize(axs[0])

        self.summerize_edges()
        segments = self.get_segments()
        if debug:
            print(segments)

        # two haplotypes only, simple walk through the graph
        if len(segments) == 1 and list(segments.values())[0] == 2:
            segment_start, segment_end = list(segments.keys())[0]
            assembled_haps = self.path(segment_start, segment_end + 1)
            if make_plot:
                self.plotfig()
            return assembled_haps

        # identify simple parts of the graph with just two haplotypes
        for i, (segment_start, segment_end) in enumerate(segments):
            if segments[(segment_start, segment_end)] == 2:
                if i not in [0, len(segments) - 1]:
                    segment_start += 1
                if segment_end == self.nvar - 2:
                    segment_end += 1
                if segment_end > segment_start:
                    self.merge_edges_simple(
                        segment_start,
                        segment_end,
                    )
        if make_plot:
            self.visualize(axs[1])
        # identify more complex parts of the graph
        for i, (segment_start, segment_end) in enumerate(segments):
            if (
                segments[(segment_start, segment_end)] > 2
                and segment_end >= segment_start
            ):
                segment_end += 1
                _ = self.merge_edges(
                    segment_start,
                    segment_end,
                    debug=False,
                )
                if debug:
                    print(segment_start, segment_end, 10)
                    pprint(self.edge_per_position)
        if make_plot:
            self.visualize(axs[2])

        if debug:
            print("done with segments")

        nstart, nend = self.get_nstart_nend()
        if nstart == nend:
            return [a.split("-")[0] for a in self.nodes]

        # scan whole graph and merge when possible

        # generalize this special logic for future
        # if first site is far away from second, start from the second site
        # if 1 in self.edge_per_position and (
        #    self.edge_info[0] == [] or min(self.edge_info[0]) < 5
        # ):
        #    nstart = 1
        while nstart < nend - 1:
            if debug:
                print(f"scanning {nstart}-{nend}")
            success, final_haps, _, _ = self.subregion_assembly(
                nstart, nend, debug=debug
            )
            current_haps = self.get_hap_by_pos(nstart)
            pos = nstart + len(final_haps[0])
            block_ending_pos, _ = self.get_previous_pos(pos)
            if len(current_haps[0]) != len(final_haps[0]):
                self.rm_add_edges(nstart, block_ending_pos, final_haps)
            nstart = pos
            if success:
                break

        if make_plot:
            self.visualize(axs[3])

        if debug:
            print("done with scanning...")
            pprint(self.edge_per_position)

        # merge blocks by size
        self.merge_blocks_by_size(debug=debug)
        if make_plot:
            self.visualize(axs[4])
            self.plotfig()

        # final merge
        nstart, nend = self.get_nstart_nend()
        if nstart == nend:
            return [a.split("-")[0] for a in self.nodes]
        if debug:
            print(f"running final {nstart}-{nend}")
            pprint(self.nodes)
        success, final_haps, _, _ = self.subregion_assembly(
            nstart, nend, allow_x=True, debug=debug
        )

        return final_haps

    def order_blocks_by_size(self):
        """When merging blocks, longer ones will be done first"""
        hap_lens = set()
        for node in self.nodes:
            hap, pos = node.split("-")
            pos = int(pos)
            hap_len = len(hap)
            hap_lens.add((pos, hap_len))
        hap_lens = sorted(hap_lens, key=lambda x: x[0])
        block_order = []
        for i, node in enumerate(hap_lens):
            if i < len(hap_lens) - 1:
                len1 = node[1]
                next_node = hap_lens[i + 1]
                len2 = next_node[1]
                block_order.append(
                    (node[0], next_node[0], min(len1, len2), max(len1, len2))
                )
        block_order = sorted(block_order, key=lambda x: (0 - x[2], 0 - x[3]))
        return block_order

    def merge_blocks_by_size(self, debug=False):
        """Merge blocks, ordered by size"""
        while self.edge_per_position != {}:
            block_order = self.order_blocks_by_size()
            if len(block_order) == 1:
                break
            if debug:
                print("priority", block_order)
            i = 0
            for pos1, pos2, _, _ in block_order:
                i += 1
                success, new_haps, _, _ = self.subregion_assembly(
                    pos1, pos2, debug=debug
                )
                if success is True:
                    if debug:
                        print(f"merging blocks {pos1}-{pos2}")
                    self.rm_add_edges(pos1, pos2, new_haps)
                    break
            if i == len(block_order) and success is False:
                break

    def run(self, debug=False, make_plot=False):
        """Run the whole haplotype phasing process"""
        self.initialize_graph(debug=debug)
        final_haps = self.assemble_haplotypes(debug=debug, make_plot=make_plot)
        if debug:
            print("assembled_haps are", final_haps)

        final_haps = self.filter_low_support_haps(final_haps, min_count=4, debug=debug)
        highest_cn, main_haps_candidates = self.get_highest_cn(final_haps)

        # extend all possible haps
        extended_main_hap_candidates = []
        for main_haps_candidate in main_haps_candidates:
            extended = self.extend_pivot_blocks(main_haps_candidate, debug=debug)
            extended_main_hap_candidates.append(extended)
        scores = []
        for extended_main_hap_candidate in extended_main_hap_candidates:
            pivot_ones = [
                a for a in extended_main_hap_candidate if a[self.pivot_pos] != "x"
            ]
            nhap = len(pivot_ones)
            total_length = sum([len(a.strip("x")) for a in pivot_ones])

            represented_haps = []
            for hap1 in final_haps:
                mismatches = []
                for hap2 in pivot_ones:
                    match, mismatch, extend = VariantGraph.compare_two_haps(hap1, hap2)
                    mismatches.append(mismatch)
                if 0 in mismatches:
                    represented_haps.append(hap1)
            num_represented_haps = len(represented_haps)

            scores.append(((nhap, num_represented_haps, total_length), pivot_ones))
        sort_candidates = sorted(
            scores,
            key=lambda x: x[0],
            reverse=True,
        )
        main_haps = sort_candidates[0][1]
        return main_haps, final_haps, highest_cn



# paraphase
# Author: Xiao Chen <xchen@pacificbiosciences.com>


import numpy as np
import pysam


class GenomeDepth:
    def __init__(self, bam, config):
        self.bam = bam
        self.config = config
        self._bamh = pysam.AlignmentFile(bam, "rb")
        self.mdepth = None
        self.mad = None

    def get_genome_depth(self):
        depth = []
        with open(self.config["data"]["depth_region"]) as f:
            for line in f:
                at = line.split()
                nchr = at[0]
                pos1 = int(at[1])
                site_depth = 0
                for pos in [pos1, pos1 + 1600]:
                    for pileupcolumn in self._bamh.pileup(
                        nchr, pos - 1, pos, truncate=True
                    ):
                        site_depth = pileupcolumn.get_num_aligned()
                depth.append(site_depth)
        self.mdepth = np.median(depth)
        self.mad = np.median([abs(a - self.mdepth) for a in depth]) / self.mdepth

    def call(self):
        self.get_genome_depth()
        self._bamh.close()
        return self.mdepth, self.mad



# paraphase
# Author: Xiao Chen <xchen@pacificbiosciences.com>


import pysam
from pprint import pprint
import os
import numpy as np
import copy
from collections import Counter
import re
from scipy.stats import poisson
from .haplotype_assembler import VariantGraph


class Phaser:

    clip_5p = r"^\d+S|^\d+H"
    clip_3p = r"\d+S$|\d+H$"
    deletion = r"\d+D"

    def __init__(self, sample_id, outdir, config, wgs_depth=None):
        self.outdir = outdir
        self.sample_id = sample_id
        self.bam = os.path.join(outdir, self.sample_id + "_realigned.bam")
        if os.path.exists(self.bam) is False:
            raise Exception(f"File {self.bam} not found.")
        self._bamh = pysam.AlignmentFile(self.bam, "rb")
        self.homopolymer_file = config["data"]["homopolymer"]
        self.homopolymer_sites = {}
        self.het_sites = []  # for phasing
        self.het_no_phasing = []
        self.homo_sites = []
        self.candidate_pos = set()
        self.mdepth = wgs_depth
        self.nchr = config["coordinates"]["hg38"]["nchr"]
        self.ref = config["data"]["reference"]
        self._refh = pysam.FastaFile(self.ref)
        self.left_boundary = config["coordinates"]["hg38"]["left_boundary"]
        self.right_boundary = config["coordinates"]["hg38"]["right_boundary"]
        self.pivot_site = config["coordinates"]["hg38"]["pivot_site"]
        self.nchr_old = config["coordinates"]["hg38"]["nchr_old"]
        self.offset = int(self.nchr_old.split("_")[1]) - 1

    def get_homopolymer(self):
        """Parse the homopolymer site file"""
        with open(self.homopolymer_file) as f:
            for line in f:
                at = line.split()
                self.homopolymer_sites.setdefault(int(at[1]), at[-1])

    @staticmethod
    def depth_prob(nread, haploid_depth):
        """Find probability of cn state based on depth"""
        prob = []
        for i in range(4):
            depthexpected = (i + 1) * haploid_depth
            pmf = poisson.pmf(nread, depthexpected)
            prob.append(pmf)
        sum_prob = sum(prob)
        if sum_prob == 0:
            return None
        post_prob = [float(a) / float(sum_prob) for a in prob]
        return post_prob

    @staticmethod
    def check_del(read, del_size):
        """Find reads having the 6.3kb deletion in its cigar string"""
        del_len = [int(a[:-1]) for a in re.findall(Phaser.deletion, read.cigarstring)]
        if del_len != [] and abs(max(del_len) - del_size) < 50:
            return True
        return False

    def get_long_del_reads(
        self,
        p3_pos1,
        p3_pos2,
        p5_pos1,
        p5_pos2,
        del_size,
        min_clip_len=300,
        min_extend=1000,
    ):
        """
        Find reads having big deletions. (Improve to general SVs for future)
        Could be softclipped at either side or have the deletion in cigar.
        Parameters:
            min_clip_len (int): minimum length for the soft-clip
        Returns: fully-spanning reads (set), partially spanning reads (set)
        """
        bamh = self._bamh
        p5_reads = set()
        p3_reads = set()
        del_reads = set()
        # 3 prime clip
        pos1 = p3_pos1
        pos2 = p3_pos2
        reference_start_cutoff = pos1 - min_extend
        for read in bamh.fetch(self.nchr, pos1, pos2):
            find_clip_3p = re.findall(self.clip_3p, read.cigarstring)
            if find_clip_3p != [] and pos1 < read.reference_end < pos2:
                if (
                    int(find_clip_3p[0][:-1]) >= min_clip_len
                    and read.reference_start < reference_start_cutoff
                ):
                    p3_reads.add(read.query_name)
            if self.check_del(read, del_size):
                del_reads.add(read.query_name)
        # 5 prime clip
        pos1 = p5_pos1
        pos2 = p5_pos2
        reference_end_cutoff = pos2 + min_extend
        for read in bamh.fetch(self.nchr, pos1, pos2):
            find_clip_5p = re.findall(self.clip_5p, read.cigarstring)
            if find_clip_5p != [] and pos1 < read.reference_start < pos2:
                if (
                    int(find_clip_5p[0][:-1]) >= min_clip_len
                    and read.reference_end > reference_end_cutoff
                ):
                    p5_reads.add(read.query_name)
            if self.check_del(read, del_size):
                del_reads.add(read.query_name)
        if del_reads != set() or (p3_reads != set() and p5_reads != set()):
            return (
                del_reads.union(p3_reads.intersection(p5_reads)),
                del_reads.union(p3_reads).union(p5_reads),
            )
        return set(), set()

    def get_pivot_site_index(self):
        """Return the index of the pivot site in list of het sites"""
        positions = [int(site.split("_")[0]) for site in self.het_sites]
        if self.pivot_site in positions:
            return positions.index(self.pivot_site), True
        return -1, False

    def get_haplotypes_from_reads(self, het_sites, exclude_reads=[], min_mapq=5):
        """
        Go through reads and get bases at sites of interest
        Returns:
            read_haps (dict of str:list): collapse each read into just the positions
            of interest. 1 corresponds to ref, 2 corresponds to alt
        """
        read_haps = {}
        nvar = len(het_sites)
        for dsnp_index, allele_site in enumerate(het_sites):
            snp_position_gene1, allele1, allele2, *at = allele_site.split("_")
            snp_position = int(snp_position_gene1)
            for pileupcolumn in self._bamh.pileup(
                self.nchr,
                snp_position - 1,
                snp_position,
                truncate=True,
                min_base_quality=30,
            ):
                for read in pileupcolumn.pileups:
                    read_name = read.alignment.query_name
                    if (
                        not read.is_del
                        and not read.is_refskip
                        and not read.alignment.is_secondary
                        and read.alignment.mapping_quality >= min_mapq
                        and read_name not in exclude_reads
                    ):
                        read_seq = read.alignment.query_sequence
                        start_pos = read.query_position
                        end_pos = start_pos + 1
                        if end_pos < len(read_seq):
                            hap = read_seq[start_pos:end_pos]
                            if read_name not in read_haps:
                                read_haps.setdefault(read_name, ["x"] * nvar)
                            if hap.upper() == allele1.upper():
                                read_haps[read_name][dsnp_index] = "1"
                            elif hap.upper() == allele2.upper():
                                read_haps[read_name][dsnp_index] = "2"
        return read_haps

    def allow_del_bases(self, pos):
        return False

    def process_indel(self, pos, ref_seq, var_seq):
        """Translate pysam indel seq into real sequence"""
        if "+" in var_seq:
            ins_base = var_seq.split(re.findall(r"\+\d+", var_seq)[0])[1]
            indel_size = len(ins_base)
            var_seq = ref_seq + ins_base
        else:
            del_len = int(re.findall(r"\-\d+", var_seq)[0][1:])
            indel_size = del_len
            var_seq = ref_seq
            offset_pos = pos - self.offset
            ref_seq = self._refh.fetch(
                self.nchr_old,
                offset_pos - 1,
                offset_pos + del_len,
            )
        return ref_seq, var_seq, indel_size

    def get_candidate_pos(self, regions_to_check=[], min_read_support=5, min_vaf=0.11):
        """
        Get all polymorphic sites in the region, update self.candidate_pos
        """
        bamh = self._bamh
        pileups_raw = {}
        for pileupcolumn in bamh.pileup(
            self.nchr,
            self.left_boundary,
            self.right_boundary,
            truncate=True,
        ):
            pos = pileupcolumn.pos + 1
            pileups_raw.setdefault(
                pos,
                [a.upper() for a in pileupcolumn.get_query_sequences(add_indels=True)],
            )
        variants = {}
        variants_no_phasing = {}
        for pos in pileups_raw:
            all_bases = pileups_raw[pos]
            total_depth = len(all_bases)
            del_bases_count = all_bases.count("*")
            # get reference base
            offset_pos = pos - self.offset
            ref_seq = self._refh.fetch(self.nchr_old, offset_pos - 1, offset_pos)

            if total_depth >= min_read_support and (
                del_bases_count < min_read_support or self.allow_del_bases(pos)
            ):
                all_bases = [a for a in all_bases if a != "*"]
                counter = Counter(all_bases)
                bases = counter.most_common(2)
                if (
                    len(counter) >= 2
                    and bases[1][1] >= min_read_support
                    and bases[1][1] / total_depth > min_vaf
                ):
                    var_seq = None
                    found_ref = ref_seq in [a[0] for a in bases]
                    for base in bases:
                        if base[0] != ref_seq:
                            var_seq = base[0]
                            break
                    if found_ref is True and var_seq is not None:
                        # SNV
                        if "-" not in var_seq and "+" not in var_seq:
                            if pos not in self.homopolymer_sites:
                                variants.setdefault(pos, (ref_seq, var_seq))
                            else:
                                prohibited_bases = self.homopolymer_sites[pos].split(
                                    ","
                                )
                                if var_seq not in prohibited_bases:
                                    if "1" in prohibited_bases:
                                        variants.setdefault(pos, (ref_seq, var_seq))
                                    else:
                                        variants_no_phasing.setdefault(
                                            pos, (ref_seq, var_seq)
                                        )
                        # indels
                        elif pos not in self.homopolymer_sites:
                            ref_seq, var_seq, indel_size = self.process_indel(
                                pos, ref_seq, var_seq
                            )
                            if indel_size < 25:
                                variants_no_phasing.setdefault(pos, (ref_seq, var_seq))
                elif len(counter) == 1 or (
                    len(counter) >= 2
                    and bases[0][1] > len(all_bases) - min_read_support
                ):
                    var_seq = bases[0][0]
                    if var_seq != ref_seq:
                        # SNV and indels
                        if "-" not in var_seq and "+" not in var_seq:
                            # "homo" sites in large deletions should be put back into het sites
                            if (
                                self.allow_del_bases(pos)
                                and del_bases_count >= min_read_support
                                and pos not in self.homopolymer_sites
                            ):
                                variants.setdefault(pos, (ref_seq, var_seq))
                            elif pos not in self.homopolymer_sites or (
                                pos in self.homopolymer_sites
                                and var_seq
                                not in self.homopolymer_sites[pos].split(",")
                            ):
                                self.homo_sites.append(f"{pos}_{ref_seq}_{var_seq}")
                        elif pos not in self.homopolymer_sites:
                            ref_seq, var_seq, indel_size = self.process_indel(
                                pos, ref_seq, var_seq
                            )
                            if indel_size < 25:
                                self.homo_sites.append(f"{pos}_{ref_seq}_{var_seq}")
        # exclude variants caused by shifted softclips of the big deletions
        excluded_variants = []
        for region in regions_to_check:
            var_to_check = [a for a in variants if region[0] < a < region[1]]
            excluded_variants += var_to_check
        for pos in variants:
            if pos not in excluded_variants:
                ref_seq, var_seq = variants[pos]
                self.candidate_pos.add(f"{pos}_{ref_seq}_{var_seq}")

        excluded_variants = []
        for region in regions_to_check:
            var_to_check = [a for a in variants_no_phasing if region[0] < a < region[1]]
            excluded_variants += var_to_check
        for pos in variants_no_phasing:
            if pos not in excluded_variants:
                ref_seq, var_seq = variants_no_phasing[pos]
                self.het_no_phasing.append(f"{pos}_{ref_seq}_{var_seq}")

    @staticmethod
    def simplify_read_haps(read_haps):
        """Simplify read haplotypes for output"""
        haplotypes_to_reads = {}
        reads_to_haplotypes = {}
        for read in read_haps:
            hap = read_haps[read]
            haplotypes_to_reads.setdefault("".join(hap), []).append(read)
            reads_to_haplotypes.setdefault(read, "".join(hap))
        return haplotypes_to_reads, reads_to_haplotypes

    def check_variants_in_haplotypes(self, variant):
        """
        For variants not used in phasing, check which haplotypes they are in.
        """
        dreads = {}
        var_pos, ref, alt = variant.split("_")
        var_pos = int(var_pos)
        var_size = len(alt) - len(ref)
        if var_size < 0:
            indel_N = "N" * abs(var_size)
            indel_base_in_read = f"{ref[0]}{var_size}{indel_N}"
        elif var_size > 0:
            indel_base_in_read = f"{ref[0]}+{var_size}{alt[1:]}"
        ref_base = ref[0]
        alt_base = alt[0] if var_size == 0 else indel_base_in_read
        for pileupcolumn in self._bamh.pileup(
            self.nchr,
            var_pos - 1,
            var_pos,
            truncate=True,
        ):
            read_names = pileupcolumn.get_query_names()
            bases = [
                a.upper() for a in pileupcolumn.get_query_sequences(add_indels=True)
            ]
            for i, read in enumerate(read_names):
                if bases[i] == ref_base:
                    dreads.setdefault(read, "ref")
                elif bases[i] == alt_base:
                    dreads.setdefault(read, "alt")
                else:
                    dreads.setdefault(read, ".")
        return dreads

    @staticmethod
    def get_start_end(hap):
        """get range of positions that are not x"""
        haplen = len(hap)
        for nstart, base in enumerate(hap):
            if base != "x":
                break
        for nend in reversed(range(haplen)):
            if hap[nend] != "x":
                break
        return nstart, nend

    def get_hap_variant_ranges(self, hap):
        """get boundaries of (partial) haplotypes"""
        nstart, nend = self.get_start_end(hap)
        if nstart == 0:
            nstart_previous_pos = self.left_boundary
        else:
            nstart_previous = nstart - 1
            nstart_previous_pos = int(self.het_sites[nstart_previous].split("_")[0]) + 1
        if nend == len(hap) - 1:
            nend_next_pos = self.right_boundary
        else:
            nend_next = nend + 1
            nend_next_pos = int(self.het_sites[nend_next].split("_")[0]) - 1
        return nstart_previous_pos, nend_next_pos

    def output_variants_in_haplotypes(self, haps, reads, nonunique):
        """
        Summarize all variants in each haplotype.
        Output all variants and their genotypes.
        Haplotypes are different length, so a range (boundary) is reported
        """
        het_sites = self.het_sites
        haplotype_variants = {}
        dvar = {}
        var_no_phasing = copy.deepcopy(self.het_no_phasing)
        for hap_index, hap in enumerate(haps):
            hap_name = f"hap{hap_index}"
            haplotype_variants.setdefault(hap_name, [])
        # het sites not used in phasing
        if reads != {}:
            for var in var_no_phasing:
                genotypes = []
                var_reads = self.check_variants_in_haplotypes(var)
                haps_with_variant = []
                for hap_index, hap in enumerate(haps):
                    hap_name = f"hap{hap_index}"
                    hap_reads = reads[hap]
                    hap_reads_nonunique = [a for a in nonunique if hap in nonunique[a]]
                    genotype = self.get_genotype_in_hap(
                        var_reads, hap_reads, hap_reads_nonunique
                    )
                    genotypes.append(genotype)
                    if genotype == "1":
                        haps_with_variant.append(hap_name)
                if haps_with_variant == []:
                    self.het_no_phasing.remove(var)
                else:
                    for hap_name in haps_with_variant:
                        haplotype_variants[hap_name].append(var)
                    dvar.setdefault(var, genotypes)
        # het sites and homo sites
        for hap_index, hap in enumerate(haps):
            hap_name = f"hap{hap_index}"
            for i in range(len(hap)):
                if hap[i] == "2":
                    haplotype_variants[hap_name].append(het_sites[i])
            # need some coordinate check if there is long deletion
            haplotype_variants[hap_name] += self.homo_sites

            var_nstart, var_nend = self.get_hap_variant_ranges(hap)
            var_tmp = haplotype_variants[hap_name]
            var_tmp1 = [
                a for a in var_tmp if var_nstart <= int(a.split("_")[0]) <= var_nend
            ]
            var_tmp1 = list(set(var_tmp1))
            haplotype_variants[hap_name] = sorted(
                var_tmp1,
                key=lambda x: int(x.split("_")[0]),
            )
            haplotype_variants[hap_name].append((var_nstart, var_nend))

        # summary per variant
        all_haps = haps
        nhap = len(all_haps)
        for var in self.homo_sites:
            dvar.setdefault(var, ["1"] * nhap)
        for i, var in enumerate(het_sites):
            dvar.setdefault(var, ["."] * nhap)
            for hap_index in range(len(all_haps)):
                hap = all_haps[hap_index]
                if hap[i] == "2":
                    dvar[var][hap_index] = "1"
                elif hap[i] == "1":
                    dvar[var][hap_index] = "0"

        return haplotype_variants, {
            var: "|".join(dvar[var]) for var in dict(sorted(dvar.items()))
        }

    def get_genotype_in_hap(self, var_reads, hap_reads, hap_reads_nonunique):
        """For a given variant, return its status in a haplotype"""
        hap_reads_contain_var = [
            var_reads[a] for a in hap_reads_nonunique if a in var_reads
        ]
        if len(hap_reads_contain_var) < 3:
            hap_reads_contain_var += [var_reads[a] for a in hap_reads if a in var_reads]
        if len(hap_reads_contain_var) >= 3:
            hap_reads_contain_var_counter = Counter(hap_reads_contain_var).most_common(
                2
            )
            if len(hap_reads_contain_var_counter) == 1 or hap_reads_contain_var_counter[
                1
            ][1] <= min(hap_reads_contain_var_counter[0][1] * 0.15, 2):
                if hap_reads_contain_var_counter[0][0] == "alt":
                    return "1"
                elif hap_reads_contain_var_counter[0][0] == "ref":
                    return "0"
        return "."

    @staticmethod
    def update_reads_for_deletions(
        raw_read_haps, het_sites, n1, n2, del_reads_partial, base, del_name
    ):
        """
        For reads carrying known big deletions, update read haplotype to
        reflect the deletion. This is needed for downstream phasing
        """
        pos1 = -1
        pos2 = -1
        for i, var in enumerate(het_sites):
            if int(var.split("_")[0]) > n1:
                pos1 = i
                break
        for i, var in enumerate(het_sites):
            if int(var.split("_")[0]) > n2:
                pos2 = i
                break
        if pos1 != -1 and pos2 != -1:
            if pos1 < pos2:
                for read in del_reads_partial:
                    if read in raw_read_haps:
                        hap = list(raw_read_haps[read])
                        for i in range(pos1, pos2):
                            hap[i] = base
                        raw_read_haps[read] = "".join(hap)
            elif pos1 == pos2:
                het_sites.insert(pos1, del_name)
                for read in raw_read_haps:
                    hap = list(raw_read_haps[read])
                    hap.insert(pos1, "x")
                    if read in del_reads_partial:
                        hap[pos1] = base
                    elif (
                        "x" not in hap[(pos1 - 2) : pos1]
                        and "x" not in hap[(pos1 + 1) : (pos1 + 3)]
                    ):
                        hap[pos1] = "1"
                    raw_read_haps[read] = "".join(hap)
        return raw_read_haps, het_sites

    def get_read_counts(self, uniquely_supporting_haps):
        """
        Get unique supporting read counts for each haplotype
        """
        nhap = len(uniquely_supporting_haps)
        nvar = len(self.het_sites)
        hap_bases = []
        for i in range(nhap):
            hap_bases.append([])
        j = 0
        for hap in uniquely_supporting_haps:
            reads = uniquely_supporting_haps[hap]
            for i in range(nvar):
                bases = [a[i] for a in reads]
                hap_bases[j].append(len(bases) - bases.count("x"))
            j += 1
        ranges = []
        for i in range(nvar):
            if min([hap_bases[a][i] for a in range(nhap)]) >= 5:
                for j in range(nvar):
                    if j > i:
                        if min([hap_bases[a][j] for a in range(nhap)]) < 5:
                            ranges.append([i, j])
                            break
                    if j == nvar - 1 and j > i:
                        ranges.append([i, nvar - 1])
        if ranges == []:
            return None
        longest_range = sorted(ranges, key=lambda x: x[1] - x[0], reverse=True)[0]
        mid = int((longest_range[1] + longest_range[0]) / 2)
        nstart = max(mid - 1, longest_range[0])
        nend = min(mid + 1, longest_range[1])
        if nend == nstart:
            return None

        read_count = {}
        for hap in uniquely_supporting_haps:
            reads = uniquely_supporting_haps[hap]
            lreads = [a for a in reads if a[nstart:nend] != "x" * (nend - nstart)]
            read_count.setdefault(hap, len(lreads))
        return read_count

    def phase_haps(self, raw_read_haps, debug=False):
        """
        Assemble and evaluate haplotypes
        """
        het_sites = self.het_sites
        haplotypes_to_reads, raw_read_haps = self.simplify_read_haps(raw_read_haps)

        ass_haps = []
        original_haps = []
        nvar = len(het_sites)
        hcn = 0
        if nvar == 0:
            return ([], [], 0, {}, {}, {}, None)
        elif nvar == 1:
            ass_haps = ["1", "2"]
            original_haps = ["1", "2"]
        else:
            pivot_index, _ = self.get_pivot_site_index()
            hap_graph = VariantGraph(
                raw_read_haps, pivot_index, figure_id=self.sample_id
            )
            ass_haps, original_haps, hcn = hap_graph.run(debug=debug, make_plot=debug)

        read_support = VariantGraph.match_reads_and_haplotypes(raw_read_haps, ass_haps)

        uniquely_supporting_haps = read_support.unique
        read_counts = self.get_read_counts(uniquely_supporting_haps)

        uniquely_supporting_reads = {}
        for hap in uniquely_supporting_haps:
            for read_hap in uniquely_supporting_haps[hap]:
                uniquely_supporting_reads.setdefault(hap, [])
                uniquely_supporting_reads[hap] += haplotypes_to_reads[read_hap]
        for hap in uniquely_supporting_haps:
            uniquely_supporting_reads[hap] = list(set(uniquely_supporting_reads[hap]))

        nonuniquely_supporting_reads = {}
        for read in read_support.by_read:
            num_matches = len(read_support.by_read[read])
            if num_matches > 1:
                nonuniquely_supporting_reads.setdefault(
                    read, read_support.by_read[read]
                )
        return (
            ass_haps,
            original_haps,
            hcn,
            uniquely_supporting_reads,
            nonuniquely_supporting_reads,
            raw_read_haps,
            read_counts,
        )

    def close_handle(self):
        self._bamh.close()
        self._refh.close()



# paraphase
# Author: Xiao Chen <xchen@pacificbiosciences.com>


import os
import argparse
import json
import yaml
import logging
import datetime
import shutil
import multiprocessing as mp
from functools import partial
from paraphase.genome_depth import GenomeDepth
from paraphase.prepare_bam_and_vcf import (
    BamRealigner,
    BamTagger,
    TwoGeneVcfGenerater,
)
from paraphase.smn_phaser import SmnPhaser


def process_sample(bamlist, outdir, config, dcov={}, vcf=False):
    for bam in bamlist:
        sample_id = bam.split("/")[-1].split(".")[0]
        logging.info(f"Processing sample {sample_id} at {datetime.datetime.now()}...")

        logging.info(f"Getting genome depth at {datetime.datetime.now()}...")
        gdepth = None
        if sample_id in dcov:
            gdepth = dcov[sample_id]
        if gdepth is None:
            depth = GenomeDepth(bam, config)
            gdepth, gmad = depth.call()
            if gdepth < 10 or gmad > 0.25:
                logging.warning(
                    "Due to low or highly variable genome coverage, genome coverage is not used for depth correction."
                )
                gdepth = None

        logging.info(f"Realigning reads at {datetime.datetime.now()}...")
        bam_realigner = BamRealigner(bam, outdir, config)
        bam_realigner.write_realign_bam()

        logging.info(f"Phasing haplotypes at {datetime.datetime.now()}...")
        smn_phaser = SmnPhaser(sample_id, outdir, config, gdepth)
        smn_phaser_call = smn_phaser.call()._asdict()

        logging.info(f"Tagging reads at {datetime.datetime.now()}...")
        bam_tagger = BamTagger(sample_id, outdir, config, smn_phaser_call)
        bam_tagger.write_bam(random_assign=True)

        if vcf:
            logging.info(f"Generating VCFs at {datetime.datetime.now()}...")
            vcf_generater = TwoGeneVcfGenerater(
                sample_id, outdir, config, smn_phaser_call
            )
            vcf_generater.run()

        sample_out = smn_phaser_call
        logging.info(f"Writing to json at {datetime.datetime.now()}...")
        out_json = os.path.join(outdir, sample_id + ".json")
        with open(out_json, "w") as json_output:
            json.dump(sample_out, json_output, indent=4)


def main():
    parser = argparse.ArgumentParser(
        description="paraphase: HiFi-based SMN1/SMN2 variant caller."
    )
    input_group = parser.add_mutually_exclusive_group(required=True)
    input_group.add_argument(
        "-b",
        "--bam",
        help="input BAM file",
    )
    input_group.add_argument(
        "-l",
        "--list",
        help="File listing absolute paths to input BAM files, one per line",
    )
    parser.add_argument(
        "-o",
        "--out",
        help="Output directory",
        required=True,
    )
    parser.add_argument(
        "-v",
        "--vcf",
        help="Option to run DeepVariant and produce vcf for each haplotype",
        required=False,
        action="store_true",
    )
    parser.add_argument(
        "-c",
        "--config",
        help="Optional path to config yaml file",
        required=False,
    )
    parser.add_argument(
        "-d",
        "--depth",
        help="Optional path to a file listing average depth for each sample",
        required=False,
    )
    parser.add_argument(
        "-t",
        "--threads",
        help="Optional number of threads. Only used when a list of inputs are given",
        required=False,
        type=int,
        default=1,
    )
    parser.add_argument(
        "--samtools",
        help="Optional path to samtools",
        required=False,
    )
    parser.add_argument(
        "--minimap2",
        help="Optional path to a minimap2",
        required=False,
    )
    parser.add_argument(
        "--singularity",
        help="Optional path to singularity",
        required=False,
    )

    args = parser.parse_args()
    outdir = args.out
    config_file = args.config
    logging.basicConfig(level=logging.DEBUG)

    data_path = os.path.join(os.path.dirname(__file__), "data")
    if config_file is None:
        config_file = os.path.join(data_path, "smn1", "config.yaml")

    # parse config file
    with open(config_file, "r") as f:
        try:
            config = yaml.safe_load(f)
        except yaml.YAMLError as yaml_error:
            raise Exception(f"Error reading {config_file}: \n\n{yaml_error}")

    # check third-party tools
    tools = config.get("tools")
    samtools_check = [
        a
        for a in [tools.get("samtools"), args.samtools, shutil.which("samtools")]
        if a is not None
    ]
    samtools_check2 = [a for a in samtools_check if os.path.exists(a)]
    if samtools_check2 == []:
        raise Exception("samtools is not found")
    else:
        config["tools"]["samtools"] = samtools_check2[0]

    minimap2_check = [
        a
        for a in [tools.get("minimap2"), args.minimap2, shutil.which("minimap2")]
        if a is not None
    ]
    minimap2_check2 = [a for a in minimap2_check if os.path.exists(a)]
    if minimap2_check2 == []:
        raise Exception("minimap2 is not found")
    else:
        config["tools"]["minimap2"] = minimap2_check2[0]

    if args.vcf:
        singularity_check = [
            a
            for a in [
                tools.get("singularity"),
                args.singularity,
                shutil.which("singularity"),
            ]
            if a is not None
        ]
        singularity_check2 = [a for a in singularity_check if os.path.exists(a)]
        if singularity_check2 == []:
            raise Exception("singularity is not found")
        else:
            config["tools"]["singularity"] = singularity_check2[0]

    # update paths
    gene = config.get("gene")
    data_paths = config.get("data")
    for data_entry in data_paths:
        old_data_file = data_paths[data_entry]
        if data_entry != "depth_region":
            new_data_file = os.path.join(data_path, gene, old_data_file)
        else:
            new_data_file = os.path.join(data_path, old_data_file)
        data_paths[data_entry] = new_data_file

    for data_file in list(data_paths.values()):
        if os.path.exists(data_file) is False:
            raise Exception(f"File {data_file} not found.")

    # parse depth file
    dcov = {}
    if args.depth is not None:
        with open(args.depth) as f:
            for line in f:
                split_line = line.split()
                dcov.setdefault(split_line[0], float(split_line[-1]))

    # process sample(s)
    bamlist = []
    if args.bam is not None:
        if os.path.exists(args.bam) and os.path.exists(args.bam + ".bai"):
            bamlist = [args.bam]
            process_sample(
                bamlist,
                outdir,
                config,
                dcov,
                args.vcf,
            )
        else:
            print(f"{args.bam} bam or bai file doesn't exist")
    elif args.list is not None:
        with open(args.list) as f:
            for line in f:
                bam = line[:-1]
                if os.path.exists(bam) and os.path.exists(bam + ".bai"):
                    bamlist.append(bam)
                else:
                    print(f"{bam} bam or bai file doesn't exist")

        nCores = args.threads
        process_sample_partial = partial(
            process_sample,
            outdir=outdir,
            config=config,
            dcov=dcov,
            vcf=args.vcf,
        )
        bam_groups = [bamlist[i::nCores] for i in range(nCores)]
        pool = mp.Pool(nCores)
        pool.map(process_sample_partial, bam_groups)
        pool.close()
        pool.join()
    else:
        raise Exception("No input file is given")


if __name__ == "__main__":
    main()



from setuptools import setup


def readme():
    with open("README.md") as f:
        return f.read()


setup(
    name="paraphase",
    version="0.1.0",
    description="paraphase: HiFi-based SMN1/SMN2 variant caller",
    long_description=readme(),
    url="https://github.com/PacificBiosciences/paraphase",
    author="Xiao Chen",
    author_email="xchen@pacificbiosciences.com",
    license="BSD-3-Clause-Clear",
    packages=["paraphase"],
    package_data={"paraphase": ["data/*", "data/smn1/*"]},
    install_requires=["pysam", "numpy", "scipy", "networkx", "matplotlib"],
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
    entry_points={"console_scripts": ["paraphase=paraphase.paraphase:main"]},
    include_package_data=True,
    long_description_content_type="text/markdown",
)



# paraphase
# Author: Xiao Chen <xchen@pacificbiosciences.com>


import os
import pysam
import subprocess
import random
import gzip


class BamRealigner:
    """
    Extract and realign reads to region of interest
    """

    min_mapq = 50
    min_aln = 400

    def __init__(self, bam, outdir, config):
        self.bam = bam
        self.outdir = outdir
        self.nchr = config["coordinates"]["hg38"]["nchr"]
        self.ref = config["data"]["reference"]
        self.nchr_old = config["coordinates"]["hg38"]["nchr_old"]
        self.offset = int(self.nchr_old.split("_")[1]) - 1
        self.nchr_length = config["coordinates"]["hg38"]["nchr_length"]
        self.extract_region1 = config["coordinates"]["hg38"]["extract_region1"]
        self.extract_region2 = config["coordinates"]["hg38"]["extract_region2"]
        self.samtools = config["tools"]["samtools"]
        self.minimap2 = config["tools"]["minimap2"]
        self._bamh = pysam.AlignmentFile(bam, "rb")
        self.sample_id = bam.split("/")[-1].split(".")[0]
        self.realign_bam = os.path.join(
            self.outdir, self.sample_id + "_realigned_old.bam"
        )
        self.realign_out_bam = os.path.join(
            self.outdir, self.sample_id + "_realigned.bam"
        )

    def write_realign_bam(self):
        """
        Realign reads to region of interest and output a tagged bam for visualization
        """
        realign_cmd = (
            f"{self.samtools} view -F 0x100 -F 0x200 -F 0x800 {self.bam} {self.extract_region1} {self.extract_region2} | "
            + f'awk \'BEGIN {{FS="\\t"}} {{print "@" $1 "\\n" $10 "\\n+\\n" $11}}\''
            + f" | {self.minimap2} -a -x map-pb {self.ref} - | {self.samtools} view -b | {self.samtools} sort > {self.realign_bam}"
        )
        result = subprocess.run(realign_cmd, capture_output=True, text=True, shell=True)
        result.check_returncode()
        if os.path.exists(self.realign_bam) is False:
            raise Exception("Realigned bam does not exist.")

        pysam.index(self.realign_bam)
        realign_bamh = pysam.AlignmentFile(self.realign_bam, "rb")
        realign_out_bamh = pysam.AlignmentFile(
            self.realign_out_bam,
            "wb",
            reference_names=[self.nchr],
            reference_lengths=[self.nchr_length],
        )
        for read in realign_bamh.fetch(self.nchr_old):
            if (
                read.mapping_quality >= self.min_mapq
                and read.query_alignment_length >= self.min_aln
            ):
                read.reference_start += self.offset
                ltags = read.tags
                new_ltags = []
                for tag in ltags:
                    if tag[0] != "SA":
                        new_ltags.append(tag)
                    else:
                        # update the SA tag
                        new_sa = []
                        for sa_item in tag[1].split(";"):
                            if sa_item != "":
                                at = sa_item.split(",")
                                at[0] = self.nchr
                                tmp_pos = int(at[1])
                                at[1] = str(tmp_pos + self.offset)
                                mapq = int(at[4])
                                if mapq >= self.min_mapq:
                                    new_at = ",".join(at)
                                    new_sa.append(new_at)
                        if new_sa != []:
                            new_sa.append("")
                            new_ltags.append(("SA", ";".join(new_sa)))
                read.tags = new_ltags
                realign_out_bamh.write(read)
        realign_bamh.close()
        realign_out_bamh.close()
        pysam.index(self.realign_out_bam)
        os.remove(self.realign_bam)
        os.remove(self.realign_bam + ".bai")


class BamTagger:
    """
    Add HP tags to bam
    """

    read_color = "166,206,227"

    def __init__(self, sample_id, outdir, config, call_sum):
        self.sample_id = sample_id
        self.outdir = outdir
        self.call_sum = call_sum
        self.bam = os.path.join(outdir, self.sample_id + "_realigned.bam")
        self.nchr = config["coordinates"]["hg38"]["nchr"]
        self._bamh = pysam.AlignmentFile(self.bam, "rb")
        self.tmp_bam = os.path.join(self.outdir, self.sample_id + "_tmp.bam")
        self.tagged_realigned_bam = os.path.join(
            self.outdir, self.sample_id + "_realigned_tagged.bam"
        )

    def add_tag_to_read(
        self,
        read,
        hp_keys,
        reads_to_tag,
        nonunique,
        random_assign=False,
    ):
        """
        Add HP tag to each read.
        Option to randomly assign a read to a possible haplotype when
        nonuniquelly matching. The randomly assigned reads are in gray
        while unique reads are in blue.
        """
        hp_found = False
        for hap, hap_name in hp_keys.items():
            if read.qname in reads_to_tag[hap]:
                read.set_tag("HP", hap_name, "Z")
                if random_assign:
                    read.set_tag("YC", self.read_color, "Z")
                hp_found = True
        if hp_found is False:
            if random_assign is False:
                read.set_tag("HP", "Unassigned", "Z")
            else:
                if read.qname in nonunique:
                    possible_haps = nonunique[read.qname]
                    random_hap = possible_haps[
                        random.randint(0, len(possible_haps) - 1)
                    ]
                    if random_hap in hp_keys:
                        read.set_tag("HP", hp_keys[random_hap], "Z")
                else:
                    read.set_tag("HP", "Unassigned", "Z")
        return read

    def write_bam(self, random_assign=False):
        """
        Write HP tags in output bams.
        """
        call_sum = self.call_sum
        out_bamh = pysam.AlignmentFile(self.tmp_bam, "wb", template=self._bamh)
        unique_reads = call_sum.get("unique_supporting_reads")
        nonunique_reads = call_sum.get("nonunique_supporting_reads")
        if nonunique_reads is None:
            nonunique_reads = {}
        hp_keys = call_sum.get("final_haplotypes")

        for read in self._bamh.fetch(self.nchr):
            if read.is_secondary is False:  # and read.is_supplementary is False:
                read = self.add_tag_to_read(
                    read,
                    hp_keys,
                    unique_reads,
                    nonunique_reads,
                    random_assign=random_assign,
                )
                out_bamh.write(read)
        out_bamh.close()
        self._bamh.close()
        pysam.sort("-o", self.tagged_realigned_bam, self.tmp_bam)
        pysam.index(self.tagged_realigned_bam)
        os.remove(self.tmp_bam)
        os.remove(self.bam)
        os.remove(self.bam + ".bai")


class VcfGenerater:
    """
    Run DeepVariant for each haplotype and generate individual/merged vcfs
    """

    search_range = 200

    def __init__(self, sample_id, outdir, config, call_sum):
        self.sample_id = sample_id
        self.outdir = outdir
        self.call_sum = call_sum
        self.bam = os.path.join(outdir, self.sample_id + "_realigned_tagged.bam")
        self.nchr = config["coordinates"]["hg38"]["nchr"]
        self.nchr_old = config["coordinates"]["hg38"]["nchr_old"]
        nstart = int(self.nchr_old.split("_")[1])
        nend = int(self.nchr_old.split("_")[2])
        self.offset = nstart - 1
        self.vcf_region = f"1-{nstart - nend}"
        if "vcf_region" in config["coordinates"]["hg38"]:
            self.vcf_region = config["coordinates"]["hg38"]["vcf_region"]
        self.nchr_length = config["coordinates"]["hg38"]["nchr_length"]
        self.ref = config["data"]["reference"]
        self.samtools = config["tools"]["samtools"]
        self.minimap2 = config["tools"]["minimap2"]
        self.singularity = config["tools"]["singularity"]
        self.match = {}

    def get_range_in_other_gene(self, pos):
        """
        Find the correponding coordinates in the other gene
        """
        if pos in self.match:
            return self.match[pos]
        for i in range(self.search_range):
            new_pos = pos + i
            if new_pos in self.match:
                return self.match[new_pos]
        return None

    def filter_gt12(self, at):
        """
        Filter variants in the scenario of GT 1/2
        """
        alt = at[4].split(",")
        info = at[-1].split(":")
        vaf = info[4].split(",")
        vaf_float = [float(a) for a in vaf]
        variant_picked_index = vaf_float.index(max(vaf_float))
        new_alt = alt[variant_picked_index]
        new_info = [
            "1/1",
            info[1],
            info[2],
            ",".join(["0", info[3].split(",")[variant_picked_index + 1]]),
            vaf[variant_picked_index],
            ",".join(info[-1].split(",")[:2] + ["0"]),
        ]
        at[4] = new_alt
        at[-1] = ":".join(new_info)
        return at

    def filter_vcf(self, hap_vcf, hap_vcf_out, hap_range, offset):
        """
        Filter vcf and return to standard coordinate
        """
        variants = []
        with gzip.open(hap_vcf) as f, open(hap_vcf_out, "w") as fout:
            for line in f:
                line = line.decode("utf8")
                if line[0] == "#":
                    if line.startswith("##contig="):
                        fout.write(
                            f"##contig=<ID={self.nchr},length={self.nchr_length}>\n"
                        )
                    else:
                        fout.write(line)
                else:
                    at = line.split()
                    gt = at[-1].split(":")[0]
                    pos = int(at[1]) + offset
                    if (
                        at[6] == "PASS"
                        and hap_range[0] < pos < hap_range[1]
                        and "0" not in gt
                    ):
                        if gt != "1/1":
                            at = self.filter_gt12(at)
                        at[0] = self.nchr
                        at[1] = str(pos)
                        fout.write("\t".join(at) + "\n")

                        ad = at[-1].split(":")[3].split(",")[1]
                        qual = float(at[5])
                        var_name = f"{pos}_{at[3]}_{at[4]}"
                        variants.append([var_name, ad, qual])
        return variants

    def merge_vcf(self, vars_ranges):
        """
        Merge vcfs from multiple haplotypes. Very rough merging.
        """
        merged_vcf = os.path.join(self.outdir, self.sample_id + f"_variants.vcf")
        with open(merged_vcf, "w") as fout:
            fout.write("##fileformat=VCFv4.2\n")
            fout.write('##FILTER=<ID=PASS,Description="All filters passed">\n')
            fout.write('##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">\n')
            fout.write(
                '##FORMAT=<ID=AD,Number=R,Type=Integer,Description="Read depth for each allele">\n'
            )
            fout.write(f"##contig=<ID={self.nchr},length={self.nchr_length}>\n")
            header = [
                "#CHROM",
                "POS",
                "ID",
                "REF",
                "ALT",
                "QUAL",
                "FILTER",
                "INFO",
                "FORMAT",
                "default",
            ]
            fout.write("\t".join(header) + "\n")
            for vars, ranges in vars_ranges:
                vars = dict(sorted(vars.items()))
                for var in vars:
                    pos, ref, alt = var.split("_")
                    pos = int(pos)
                    call_info = vars[var]
                    merge_gt = []
                    merge_ad = []
                    merge_qual = []
                    for i, each_call in enumerate(call_info):
                        hap_range = ranges[i]
                        if each_call is None:
                            if hap_range[0] < pos < hap_range[1] or None in hap_range:
                                merge_gt.append("0")
                                merge_ad.append("0")
                            else:
                                merge_gt.append(".")
                                merge_ad.append(".")
                        else:
                            merge_gt.append("1")
                            merge_ad.append(str(each_call[0]))
                            merge_qual.append(each_call[1])

                    merged_entry = [
                        self.nchr,
                        str(pos),
                        ".",
                        ref,
                        alt,
                        str(min(merge_qual)),
                        "PASS",
                        ".",
                        "GT:AD",
                        "/".join(merge_gt) + ":" + ",".join(merge_ad),
                    ]
                    fout.write("\t".join(merged_entry) + "\n")

    def run_step(
        self,
        final_haps,
        haplotype_details,
        ref_seq,
        ref_name,
        offset,
        match_range=False,
    ):
        """
        Process one haplotype
        """
        vars = {}
        ranges = []
        two_cp_haplotypes = self.call_sum.get("two_copy_haplotypes")
        nhap = len(final_haps)
        for i, hap_name in enumerate(final_haps.values()):
            hap_range = haplotype_details[hap_name][-2]
            if match_range:
                hap_range = [
                    self.get_range_in_other_gene(hap_range[0]),
                    self.get_range_in_other_gene(hap_range[1]),
                ]
            ranges.append(hap_range)
            hap_bam = os.path.join(self.outdir, self.sample_id + f"_{hap_name}.bam")
            realign_cmd = (
                f"{self.samtools} view -d HP:{hap_name} {self.bam} |"
                + f'awk \'BEGIN {{FS="\\t"}} {{print "@" $1 "\\n" $10 "\\n+\\n" $11}}\''
                + f" | {self.minimap2} -a -x map-pb {ref_seq} - | {self.samtools} view -b | {self.samtools} sort > {hap_bam}"
            )
            result = subprocess.run(
                realign_cmd, capture_output=True, text=True, shell=True
            )
            result.check_returncode()
            pysam.index(hap_bam)

            hap_vcf = os.path.join(
                self.outdir, self.sample_id + f"_{hap_name}_prefilter.vcf.gz"
            )
            vcf_cmd = (
                f"{self.singularity} exec --bind /usr/lib/locale/ docker://google/deepvariant:1.3.0 "
                + f"/opt/deepvariant/bin/run_deepvariant --model_type PACBIO --ref {ref_seq}  --reads {hap_bam} "
                + f"--output_vcf {hap_vcf} --num_shards 1 --regions {ref_name}:{self.vcf_region}"
            )
            result = subprocess.run(vcf_cmd, capture_output=True, text=True, shell=True)
            result.check_returncode()

            # fileter vcf
            hap_vcf_out = os.path.join(self.outdir, self.sample_id + f"_{hap_name}.vcf")
            variants_called = self.filter_vcf(hap_vcf, hap_vcf_out, hap_range, offset)

            for var_name, ad, qual in variants_called:
                vars.setdefault(var_name, [None] * nhap)
                vars[var_name][i] = [ad, qual]
        # insert info for two-copy haplotype
        if two_cp_haplotypes != []:
            for i, hap_name in enumerate(final_haps.values()):
                if hap_name in two_cp_haplotypes:
                    for var in vars:
                        var_call = vars[var][i]
                        vars[var].insert(i, var_call)
                    range_call = ranges[i]
                    ranges.insert(i, range_call)
        return vars, ranges

    def run(self):
        """Process haplotypes one by one"""
        call_sum = self.call_sum
        vars, ranges = self.run_step(
            call_sum["final_haplotypes"],
            call_sum["haplotype_details"],
            self.ref,
            self.nchr_old,
            self.offset,
        )
        self.merge_vcf([(vars, ranges)])


class TwoGeneVcfGenerater(VcfGenerater):
    """
    Make vcf for two-gene scenario
    """

    def __init__(self, sample_id, outdir, config, call_sum):
        VcfGenerater.__init__(self, sample_id, outdir, config, call_sum)
        self.nchr_old_gene2 = config["coordinates"]["hg38"]["nchr_old_smn2"]
        self.offset_gene2 = int(self.nchr_old_gene2.split("_")[1]) - 1
        self.ref_gene2 = config["data"]["reference_smn2"]
        self.position_match = config["data"]["smn_match"]
        with open(self.position_match) as f:
            for line in f:
                at = line.split()
                self.match.setdefault(int(at[1]), int(at[3]))

    def run(self):
        """
        Process haplotypes one by one. Realign to different ref sequence
        in this two-gene scenario
        """
        call_sum = self.call_sum
        vars_smn1, ranges_smn1 = self.run_step(
            call_sum["smn1_haplotypes"],
            call_sum["haplotype_details"],
            self.ref,
            self.nchr_old,
            self.offset,
        )
        vars_smn2, ranges_smn2 = self.run_step(
            call_sum["smn2_haplotypes"],
            call_sum["haplotype_details"],
            self.ref_gene2,
            self.nchr_old_gene2,
            self.offset_gene2,
            match_range=True,
        )
        self.merge_vcf([(vars_smn2, ranges_smn2), (vars_smn1, ranges_smn1)])
